-- AUTHOR: Paranoidi

-- $LastChangedDate: 2005-10-13 20:53:15 +0300 (Thu, 13 Oct 2005) $
-- $LastChangedBy: marko $
-- $Rev: 139 $

-- integration to myAddons

SmartAssistDetails = {
		name = "SmartAssist",
		description = "SmartAssist is an addon which improves default assisting system in groups.",
		version = "1.1.0",
		releaseDate = string.sub("$LastChangedDate: 2005-10-13 20:53:15 +0300 (Thu, 13 Oct 2005) $", 18, 28), -- note: svn keyword
		author = "paranoidi",
		email = "paranoidi@gmx.net",
		category = MYADDONS_CATEGORY_COMBAT,
		frame = "SmartAssistFrame",
		optionsframe = "SAOptionsFrame"
};

-- some global variables

ASSIST_EMOTES = { };
PREVIOUS_ASSISTS = {};

SMARTASSIST_CORE_REV = tonumber(string.sub("$Rev: 139 $", 7, -3)); -- note: svn keyword
BINDING_HEADER_SMARTASSIST = "SmartAssist";
BINDING_NAME_SASET = "Set puller";
BINDING_NAME_SAASSIST = "Assist";

-- options
SA_OPTIONS_WRAPPER = {};
SA_OPTIONS = nil;

-- used to disable sounds and detect target changes
SA_RUNNING = false;
PASSIVE_WARN_FLAG = false;


StaticPopupDialogs["SA_NO_BINDINGS"] = {
	text = TEXT("You must bind SmartAssist key before being able to use this addon! See readme.txt if you have problems."),
	button1 = TEXT(ACCEPT),
	OnAccept = function()
	end,
	timeout = 0,
	whileDead = 1,
};

StaticPopupDialogs["SA_AUTOCONF"] = {
	text = TEXT("Auto configure Smart Actions?"),
	button1 = TEXT(ACCEPT),
	button2 = TEXT(CANCEL),
	OnAccept = function()
		SA_OPTIONS.AutoLearn = true;
		SA_AutoConfigureSmartActions();
		printInfo("You should see SmartAssist options and verify detected settings. Use command /sa to access configuration.");
	end,
	timeout = 0,
	whileDead = 1,
};

function SA_OnLoad()
	SLASH_SMARTASSIST1 = "/smartassist";
	SLASH_SMARTASSIST2 = "/sa";
	SlashCmdList["SMARTASSIST"] = function(msg)
		SA_Command(msg);
	end

	-- add item to every possible context menu	
	table.insert(UnitPopupMenus["PARTY"], "PULLER");
	table.insert(UnitPopupMenus["PLAYER"], "PULLER");
	table.insert(UnitPopupMenus["RAID"], "PULLER");
	
	UnitPopupButtons["PULLER"] = { text = TEXT("Set as puller"), dist = 0 };
end

function SA_InitOption(option, value)
	if (SA_OPTIONS[option]==nil) then SA_OPTIONS[option]=value; end;
end

function SA_GetAccountID()
	return UnitName("player").." of "..GetCVar("realmName");
end

-- event variables loaded
function SA_Init() 
	-- Register the addon in myAddOns
	if(myAddOnsFrame_Register) then
		myAddOnsFrame_Register(SmartAssistDetails);
	end

	printInfo("SmartAssist "..SmartAssistDetails.version.." loaded.")

	-- load realm & char specific configs
	local id = SA_GetAccountID();
	if (SA_OPTIONS_WRAPPER[id]==nil) then
		printInfo("SmartAssist created new default settings for "..id);
		SA_OPTIONS_WRAPPER[id] = {};
	end
	SA_OPTIONS = SA_OPTIONS_WRAPPER[id];
		
	-- set default values to option variables if they do not exist

	-- before rev 66 we had no stored rev number so we need to make sure it's in such cases atleast zero
	SA_InitOption("Rev", 0);

	-- special cases where we need to reset variables when upgrading the addon from older version		
	--if (SA_OPTIONS.Rev<66) then
	--end
	
	-- basic options
	SA_InitOption("VisualWarning", true);
	SA_InitOption("AudioWarning", false);
	SA_InitOption("AssistOnEmote", true);
	SA_InitOption("PriorizeHealth", true);
	SA_InitOption("PriorizeHealthValue", 35);
	SA_InitOption("FallbackTargetNearest", true);
	SA_InitOption("ShowAvailable", false);
	-- advanced
	SA_InitOption("DisableTargetNearest", true);
	SA_InitOption("DisableTargetNearestValue", 15);
	SA_InitOption("AssistKeyMode", 1);
	SA_InitOption("DisableAutoCastKeyMode", 2);
	SA_InitOption("VerboseAssistToChat", false);
	SA_InitOption("VerboseAssistToSct", true);
	SA_InitOption("VerboseIncomingToSct", true);
	SA_InitOption("OutOfCombat", false);
	-- smart actions
	SA_InitOption("AssistSlot", {});
	SA_InitOption("BuffSlot", {});
	SA_InitOption("AfterBuffTargetBackMob", true);
	SA_InitOption("TriggerAssist", true);
	SA_InitOption("AutoLearn", false);
	SA_InitOption("SmoothenBuffs", false); -- experimental
	-- display auto configuration dialog on first run
	SA_InitOption("AutoConfigured", false);
	if (not SA_OPTIONS.AutoConfigured) then
		SA_OPTIONS.AutoConfigured = true;
		StaticPopup_Show("SA_AUTOCONF");
	end
		
	-- init auto cast on assist defaults for hunters
	if (SA_OPTIONS["AutoAssist"]==nil and UnitClass("player") == CLASSNAME_HUNTER) then
		SA_OPTIONS.AutoAssist = true;
		SA_OPTIONS.AutoAssistTexture = "Interface\\Icons\\INV_Weapon_Rifle_08";
		SA_OPTIONS.AutoAssistName = SPELL_AUTOSHOT;
	end
	SA_InitOption("AutoAssist", false);

	-- prefered assitance order (just a guess, got better one?)
	SA_InitOption("ClassOrder",{	CLASSNAME_WARRIOR, CLASSNAME_HUNTER, 
									CLASSNAME_ROGUE, CLASSNAME_PALADIN, 
									CLASSNAME_DRUID, CLASSNAME_SHAMAN, CLASSNAME_MAGE, 
									CLASSNAME_WARLOCK, CLASSNAME_PRIEST } );
									
	-- store revision number to configuration, this can be used to detect some special cases where we need to reset
	-- exsisting options (ie. bug in assistance order list) in new releases
	SA_OPTIONS.Rev = SMARTASSIST_CORE_REV;
	
	ASSIST_EMOTES = { EMOTE_EVERYONE_ATTACK, EMOTE_ATTACK, EMOTE_HELP };
	
	if (SA_OPTIONS.ShowAvailable) then
		SAListFrame:Show();
	else
		SAListFrame:Hide();
	end
	
end

-------------------------------------------------------------------------------
-- WARNING ICON EVENTS
-------------------------------------------------------------------------------
SA_WARN_SPEED = 0.07;
SA_WARN_REFRESH = SA_WARN_SPEED;
SA_WARN_ALPHA = 0.4;
SA_WARN_ALPHA_UP = true;

function SA_WarnOnUpdate(elapsed)
	if (not PASSIVE_WARN_FLAG) then return; end;
	SA_WARN_REFRESH = SA_WARN_REFRESH - elapsed;
	if (SA_WARN_REFRESH > 0) then
		return;
	end
	SA_WARN_REFRESH = SA_WARN_SPEED;
	
	-- todo: if target target is != nil (attacking someone) then REMOVE the icon
	if (UnitName("targettarget")~=nil) then
		SA_Debug("target target != nil .. remove the warning");
		SA_ResetWarning();
		SA_AggroedTargetMsg();
		return;
	end

	if (SA_WARN_ALPHA_UP) then
		SA_WARN_ALPHA = SA_WARN_ALPHA + 0.1;
	else
		SA_WARN_ALPHA = SA_WARN_ALPHA - 0.1;
	end
	if (SA_WARN_ALPHA<0.4) then
		SA_WARN_ALPHA_UP = true;
		SA_WARN_ALPHA = 0.5;
	end
	if (SA_WARN_ALPHA>1.0) then
		SA_WARN_ALPHA_UP = false;
		SA_WARN_ALPHA = 0.9;
	end
	SAWarningFrame:SetAlpha(SA_WARN_ALPHA);
end

-------------------------------------------------------------------------------
-- handles smart assist slash commands
-------------------------------------------------------------------------------

function SA_Command(msg)
	if (msg=="debug") then
		if (SA_OPTIONS.Debug) then
			printInfo("SmartAssist debug mode disabled");
			SA_OPTIONS.Debug = false;
		else
			printInfo("SmartAssist debug mode enabled");
			SA_OPTIONS.Debug = true;
		end
	elseif (msg=="available") then
		SA_ToggleAvailable();
	elseif (msg=="assist") then
		DoSmartAssist();
	elseif (msg=="reset") then
		printInfo("Reseting SmartAssist");
		-- resets frame positions
		SAListFrame:ClearAllPoints();
		SAListFrame:SetPoint("TOP", "UIParent", "TOP", 0, -20);
		SAListFrame:StopMovingOrSizing();
		SAWarningFrame:ClearAllPoints();
		SAWarningFrame:SetPoint("TOP", "UIParent", "TOP", 5, -25);
		SAWarningFrame:StopMovingOrSizing();
		SAOptionsFrame:ClearAllPoints();
		SAOptionsFrame:SetPoint("CENTER", "UIParent", "CENTER", 0, 0);
		SAOptionsFrame:StopMovingOrSizing();
				
		-- clears this character settings and initializes defaults
		local id = SA_GetAccountID();
		SA_OPTIONS_WRAPPER[id] = {};
		SA_Init();
	elseif (msg=="rev" or msg=="ver") then
		if (SMARTASSIST_REV) then
			printInfo("SmartAssist revision "..SMARTASSIST_REV);
		else
			printInfo("Unofficial or broken build! Revision number missing! Core rev "..SA_OPTIONS.Rev);
		end
	else
		-- open config, or warn about key bindings
		local key1, key2 = GetBindingKey("SAASSIST");
		if (key1==nil and key2==nil) then
			StaticPopup_Show("SA_NO_BINDINGS");
		else
			SAOptionsFrame:Show();
		end
	end
end


-------------------------------------------------------------------------------
-- SMART ACTIONS
-------------------------------------------------------------------------------

TT_LOG = {};

SA_BUFF_REFRESH = 0.5;
function SA_TimerUpdate(elapsed)
	SA_BUFF_REFRESH = SA_BUFF_REFRESH - elapsed;
	if (SA_BUFF_REFRESH > 0) then
		return;
	end
	SA_BUFF_REFRESH = 0.5;
	local current_target = UnitName("targettarget");
	if (current_target) then
		table.insert(TT_LOG, current_target);
		if (table.getn(TT_LOG) > 4) then
			--SA_Debug("purging #1 "..TT_LOG[1]);
			table.remove(TT_LOG, 1);
		end
	elseif (table.getn(TT_LOG) > 0) then
		--SA_Debug("purging #2 "..TT_LOG[1]);
		table.remove(TT_LOG, 1);
	end
end

function SA_EnableTimer()
	if not (SA_OPTIONS.SmoothenBuffs) then return; end;
	SA_Debug("enabling smoother timer");
	SATimerFrame:Show();
end

function SA_DisableTimer()
	SA_Debug("disabling smoother timer");
	TT_LOG = {};
	SATimerFrame:Hide();
end

-- this flag is nil when non smart action is executed
-- on buff smart action it will be "**mob**"
-- on assist smart action it will be the assisted player name
local SA_SMARTACTION = nil;

local old_UseAction = UseAction;
function UseAction(m1, m2, m3)
	SA_SMARTACTION = nil;

	local buffAction = SA_InList(SA_OPTIONS.BuffSlot, m1);
	local assistAction = SA_InList(SA_OPTIONS.AssistSlot, m1);

	-- no target selected, attack action
	if (SA_OPTIONS.TriggerAssist and UnitName("target")==nil and assistAction) then
		SA_Debug("no target, initiating assist by action!");
		FindTarget(false, false);
		-- if unit is not in combat we should not just start shooting it because target is acquired via assist, 
		-- player has not decided it
		if (not UnitAffectingCombat("target")) then
			SA_Debug("not in combat, aborting action");
			SA_ShowWarning();
			return; 
		end;
	elseif (UnitName("target")==nil and buffAction) then
		-- no target, buff action .. target player
		TargetUnit("player");
	end

	-- autolearn actions on the fly
	if (SA_OPTIONS.AutoLearn) then
		--local ps = debugprofilestart();
		if (not assistAction and not buffAction) then
			local slot_name = SA_GetSlotName(m1);
			if (AUTOCONF_ATTACKS[slot_name]) then
				printInfo("Learned new Smart Action '"..slot_name.. "' (attack)");
				-- add permanently to attacks
				table.insert(SA_OPTIONS.AssistSlot, m1);
				assistAction = true;
			elseif (AUTOCONF_BUFFS[slot_name]) then
				printInfo("Learned new Smart Action '"..slot_name.."' (buff)");
				table.insert(SA_OPTIONS.BuffSlot, m1);
				buffAction = true;
			end
		end
		--SA_Debug("scannin took "..debugprofilestop().." ms");
	end
	
	-- if trying to put buff on enemy, initiate smart action
	if (buffAction and UnitCanAttack("target", "player")) then
		SA_Debug("enemy target, buffAction = "..tostring(buffAction));
		local name = UnitName("targettarget");
		SA_Debug("targettarget name = "..tostring(name));
		
		-- Determine how old this target is from the tt log, if the target has just been changed we do not want to use it.
		-- This should fix the problems when you hit heal and target has just changed its target,
		-- in this case we buff previous target.
		
		local tt_time = 255;
		local tt_name = nil;
		for k,v in TT_LOG do
			if (v~=name) then
				tt_time = k;
				tt_name = v;
				break;
			end
		end
		SA_Debug("tt log check results, tt_time="..tostring(tt_time).." tt_name="..tostring(tt_name));
		if (tt_time<=2) then
			-- target has been chaned recently
			printInfo("Targets target has been changed recently, using previous target "..tostring(tt_name));
			name = tt_name;
		end
		if (name) then
			if (SA_OPTIONS.VerboseAssistToSct and SCT_Display) then  
				SCT_Display("Assisting "..name, {r=0,g=0.8,b=1});
			end
			TargetByName(name);
			SA_SMARTACTION = "**mob**"; -- special flag, handled in spellcast end
		end
	end
	
	-- if trying to attack friendly unit, initiate smart action
	if (assistAction and UnitIsFriend("target", "player")) then
		SA_Debug("friendly target, assistAction = "..tostring(assistAction));
		if (UnitCanAttack("player", "targettarget")) then
			SA_SMARTACTION = UnitName("target");
			SA_Debug("friendly target, assisting");
			AssistUnit("target");
		else
			-- supresses annoyance when attacking and ally has no target
			if (assistAction) then 
				SA_Debug("aborting action, supressing annoyance");
				return; 
			end
		end;
	end
	
	return old_UseAction(m1, m2, m3);
end

function SA_SpellStopped()
	if (SA_SMARTACTION ~= nil) then
		if (SA_SMARTACTION=="**mob**") then
			if (SA_OPTIONS.AfterBuffTargetBackMob) then
				SA_Debug("targetting back to previous mob");
				TargetLastEnemy();
			end
		else
		    -- does not work reliable enough
			--SA_Debug("targetting back to assisted unit");
			--TargetByName(SA_SMARTACTION);
		end
		SA_SMARTACTION = nil;
	end
end

-------------------------------------------------------------------------------
-- implements shift click assist
-- should work with ANYWHERE where you can select unit
-------------------------------------------------------------------------------

local SA_PREV_ASSIST_TIME = 0;
function SA_PlayerTargetChanged()
	-- fixes issue when we have target already (fires two events, on lost and on gain)
	if (not UnitExists("target")) then 
		SA_List_Refresh();
		return;
	end
	if ( (IsShiftKeyDown() and SA_OPTIONS.AssistKeyMode==1) or 
	     (IsControlKeyDown() and SA_OPTIONS.AssistKeyMode==2) or
	     (IsAltKeyDown() and SA_OPTIONS.AssistKeyMode==3) ) 
	then
		-- this time check (hoax) removes looping, when selecting player that has friendly taget. Without this 
		-- it goes there and starts to assist him
		-- todo: this can be done better .. if we check what's the target!
		
		local now = time();
		local dif = now - SA_PREV_ASSIST_TIME;
		SA_Debug("dif is "..dif);
		if (dif == 0) then return; end
		SA_PREV_ASSIST_TIME = now;

		--SA_Debug("target is "..tostring(UnitName("target")));
		--SA_Debug("unit exists is "..tostring(UnitExists("target")));
		if (UnitCanAssist("player", "target") and
			UnitCanAttack("player", "targettarget"))
		then
			SAMsgFrame:AddMessage("Assisting "..UnitName("target"), 0.0, 0.8, 0.0, 1, 2.5);
			AssistUnit("target");
			SA_PostAssist();
		else
			if (UnitExists("target") and not UnitIsDead("target")) then
				UIErrorsFrame:AddMessage("Unable to assist", 1.0, 0, 0, 1.0, 1.5);
			end
		end
	end
	-- if we have enemy target, enable smoother buff action target timer
	if (UnitCanAttack("player", "target")) then
		SA_EnableTimer();
	else
		SA_DisableTimer();
	end
	
	SA_List_Refresh();
end

-- credits to popup assist author, refactored a bit tough :)
local old_UnitPopup_OnClick = UnitPopup_OnClick;
function UnitPopup_OnClick()
	local index = this.value;
	local dropdownFrame = getglobal(UIDROPDOWNMENU_INIT_MENU);
	local button = UnitPopupMenus[this.owner][index];
	local unit = dropdownFrame.unit;
	local name = dropdownFrame.name;
	
	if ( button == "PULLER" ) then
		SA_Debug("popup, setting puller to name="..name.." unit="..unit);
		SA_SetPuller(unit);
	end
	
	return old_UnitPopup_OnClick();
end

-- set unit as puller
function SA_SetPuller(unit)
	if(UnitIsFriend(unit, "player")) then
		local candidates = SA_GetCandidates(true);

		-- check that current puller exists in party, if not clear
		if (not candidates[UnitName(unit)]) then
			SAMsgFrame:AddMessage("Puller must be in party / raid", 1.0, 1.0, 1.0, 1, 3);
		else
			SA_OPTIONS["puller"] = UnitName(unit);
			SAMsgFrame:AddMessage("The assigned puller is now "..SA_OPTIONS["puller"], 1.0, 1.0, 1.0, 1, 1.5);
		end
	else
		SA_OPTIONS["puller"] = nil;
		SAMsgFrame:AddMessage("Assigned puller cleared", 1.0, 1.0, 1.0, 1, 1.5);
	end
	-- just incase that option window is visible, update the text
	SA_Options_UpdatePullerText();
end

-- credits to mozz
-- this is used to disable some anying unit deselect sounds while assisting 
-- (we have to clear the selected unit in some cases)
local old_PlaySound = PlaySound;
function PlaySound(name)
    if (not SA_RUNNING) then
    	return old_PlaySound(name);
    end
    
	if (name ~= "INTERFACESOUND_LOSTTARGETUNIT" and name ~= "igCharacterNPCSelect" ) then
		--SA_Debug("playing.."..name);
		return old_PlaySound(name);
	else
		--SA_Debug("supressing.."..name);
	end
end

function DoSmartAssist()
    SA_RUNNING = true;
	FindTarget(true, false);
	SA_RUNNING = false;
	SA_PostAssist();
	SA_List_Refresh();
end

function SA_PostAssist()
	-- auto-assist, disable if certain modifier key is down
	if ( (not IsShiftKeyDown() and SA_OPTIONS.DisableAutoCastKeyMode==1) or 
	     (not IsControlKeyDown() and SA_OPTIONS.DisableAutoCastKeyMode==2) or
	     (not IsAltKeyDown() and SA_OPTIONS.DisableAutoCastKeyMode==3) ) 
	then
		if (UnitExists("target") and 
		    UnitAffectingCombat("target") and 
		    SA_OPTIONS.AutoAssist and 
		    SA_OPTIONS.AutoAssistName) 
		then
			-- local isUsable, notEnoughMana = IsUsableAction(1); -- note: bar slots !
			SA_Debug("Target already in combat, assisting with "..SA_OPTIONS.AutoAssistName);
			CastSpellByName(SA_OPTIONS.AutoAssistName); -- casts highest level always
		end
	end
	-- visual alert, if needed
	SA_ShowWarning();
end

-- receives emote events, implements assist by emote
function SA_EmoteAssist(arg1)
	if (not SA_OPTIONS.AssistOnEmote) then
		SA_Debug("Emote assist disabled");
		return;
	end
	-- todo idea: hitting assist key should go back to previous target (if possible)
	-- parse emotes
	for k,v in ASSIST_EMOTES do
		--SA_Debug("parsing k="..k.." v="..v);
		local sp, ep, name = string.find(arg1, "(%w+) "..v);
		if (name ~= nil) then
			SA_Debug("assist emote detected, name="..name);
		end
		-- if name exists, we have a match
		if name and name ~= EMOTE_OWN then
			local candidates = SA_GetCandidates(true);
			if (candidates[name]~=nil) then
				SAMsgFrame:AddMessage("Assisting "..name, 0.0, 0.8, 0.0, 1, 2.5);
				AssistUnit(candidates[name].unitId);
				return true;
			end
			SA_Debug("not assisting "..name..", not in party/raid");
		end
	end
end

-- Return the spell index for the given spell name
function SA_getSpellID(spell)
	for id = 1, 180 do 
		local spellName, subSpellName = GetSpellName(id, SpellBookFrame.bookType);
		if spellName and string.find(spellName, spell, 1, true) then 
			return id
		end
	end
	return -1;
end

-- hook to target frame, clearing your target outside SA_RUNNING will clear PREVIOUS_ASSISTS
local SA_TargetFrame_OnHide = TargetFrame_OnHide;
function TargetFrame_OnHide()
    -- disable smoothen buff timer
    SA_DisableTimer();
	--SA_Debug("Target lost")
	SA_ResetWarning();
	if (not SA_RUNNING) then
		SA_Debug("Target lost! Clearing previous assists list!")
		PREVIOUS_ASSISTS = {};
	end
	-- disable buffaction smoothen timer
	return SA_TargetFrame_OnHide();
end

function SA_UnitHealth(arg1)
	-- flag is used to reduce overheading, this gets called a lot ..
	if (PASSIVE_WARN_FLAG==false) then
		return;
	end
	if (arg1=="target") then
		SA_ResetWarning();
		-- if someone else attacked the target, show another text saying its ok to start blasting
		-- this is not idiot proof check, seems to work fine tough
		if (not UnitIsUnit("targettarget", "pet") and not UnitIsUnit("targettarget", "player")) then
			SA_Debug("Someone else has attacked our target, show notification");
			SA_AggroedTargetMsg();
		else
			SA_Debug("player has most likelly attacked the target");
			-- todo: perhaps send pull message to party?
		end
	end
	-- todo idea: we could watch other players health here and suggest assist
end

-- show warning if target is not in combat and feature is enabled
function SA_ShowWarning()
	if (not UnitAffectingCombat("target") and UnitExists("target") and 
		UnitName("targettarget")==nil and SA_OPTIONS.VisualWarning) 
	then
		PASSIVE_WARN_FLAG = true;
		SAWarningFrame:Show();
		return true;
	end
end

function SA_AggroedTargetMsg()
	SAMsgFrame:AddMessage(" ", 0, 0, 0, 1, 0.8);
	SAMsgFrame:AddMessage(" ", 0, 0, 0, 1, 0.8);
	SAMsgFrame:AddMessage("target is now in combat", 0, 0.8, 0, 1, 0.8);
end

function SA_ResetWarning()
	PASSIVE_WARN_FLAG = false;
	SAWarningFrame:Hide();
end

-- construct one candidate, return nil if unit could not be constructed (does not exists)
function SA_GetCandidate(name, i)
	local candidate = {};
	candidate["unitName"] = UnitName(name..i);
	candidate["unitId"] = name..i;
	candidate["target"] = name..i.."target";
	candidate["index"] = i;
	candidate["health"] = ceil( UnitHealth( name..i ) / UnitHealthMax( name..i ) * 100 );
	candidate["class"] = UnitClass(name..i);
	if (string.find(name, "pet") ~= nil) then
		candidate["pet"] = true;
	end
	if (candidate.unitName == nil) then 
		SA_DebugCandidate(candidate);
		return nil; 
	end
	if (candidate.class == nil) then
		SA_DebugCandidate(candidate);
		return nil; 
	end
	if (candidate.health == nil) then
		printInfo("Problem with smartassist; unit health is unknown!");
		SA_DebugCandidate(candidate);
		return nil;
	end
	return candidate;
end

-- return iteration info 
function SA_GetIterInfo()
	local mode, members = nil;
	if (GetNumRaidMembers()>0) then
		--SA_Debug("raid mode");
		mode = "raid";
		members = GetNumRaidMembers();
	elseif (GetNumPartyMembers()>0) then
		--SA_Debug("party mode");
		mode = "party";
		members = GetNumPartyMembers();
	end
	return mode, members;
end

-- return list of possible assistable units
-- if map is true then as a map where key is unitName
-- if map is false then as list
function SA_GetCandidates(map)
	local candidates = {};
	local mode, members = SA_GetIterInfo();

	-- for soloing and party, add our pet
	if (UnitExists("pet")) then
		--SA_Debug("we have pet, add it to candidates");
		local op = SA_GetCandidate("pet", "");
		if (op) then 
			if (map) then
				candidates[op.unitName] = op;
			else
				table.insert(candidates, op);
			end
		else
			SA_Debug("******* SAFECHECK 1 PAID OFF **********");
		end
	end
	
	-- if no party/raid, abort here
	if (members==nil) then return candidates; end;

	local myname = UnitName("player");

	for i = 1, members do
		local candidate = SA_GetCandidate(mode, i);
		-- do not add myself to the list
		if (candidate and candidate["unitName"]~=myname) then
			if (map) then
				candidates[candidate.unitName] = candidate;
			else
				table.insert(candidates, candidate);
			end
		
			-- add pet to list if has one
			if (UnitExists(mode.."pet"..i)) then
				local pcandidate = SA_GetCandidate(mode.."pet", i);
				if (pcandidate) then
					if (map) then
						candidates[pcandidate.unitName] = pcandidate;
					else
						table.insert(candidates, pcandidate);
					end
				else
					--printInfo("******* SAFECHECK 2 PAID OFF **********");
					-- pets which are far away do not have every detail available, atleast class is unknown
				end
			end
		else
			if (not candidate) then
				SA_Debug("******* SAFECHECK 3 PAID OFF **********");
			end
		end
	end
	
	return candidates;
end

function SA_DebugCandidates(candidates)
	for k,c in candidates do
		--SA_Debug("key="..k);
		if (c==nil) then
			SA_Debug("THERE IS NIL VALUE ON ARRAY!");
		end
		SA_DebugCandidate(c);
	end
end

function SA_DebugCandidate(c)
	SA_Debug("candidate name="..tostring(c.unitName).." | id="..tostring(c.unitId).." | health="..tostring(c.health).." | class="..tostring(c.class).." | pet="..tostring(c.pet));
end

function SA_DebugTable(t)
	SA_Debug("table length: "..SA_TableSize(t));
	for k,v in t do
		SA_Debug("table: key="..k.." value="..tostring(v));
	end
end

-- sort method for candidates
local detected_bug = false;
function SA_SortCandidate(a, b)
	-- TODO: INVESTIGATE, but HOW?
	-- this is added to debug one odd behaviour in molten core where crash occured (b was nil, crashed at priorize health)
	if (detected_bug==false and (a==nil or b==nil)) then
		printInfo("?? BUG IN SMARTASSIST:");
		local cand = SA_GetCandidates(false);
		local old_state = SA_OPTIONS.Debug;
		SA_OPTIONS.Debug = true;
		SA_DebugCandidates(cand);
		SA_OPTIONS.Debug = old_state;
		detected_bug = true;
	end
	
	-- priority health always first
	if ( (SA_OPTIONS.PriorizeHealth) and 
	     (a["health"] < SA_OPTIONS.PriorizeHealthValue or b["health"] < SA_OPTIONS.PriorizeHealthValue) ) 
	then
		return a["health"] < b["health"];
	end
	
	-- our puller should be always our first priority
	if (a["unitName"] == SA_OPTIONS["puller"]) then return true; end
	if (b["unitName"] == SA_OPTIONS["puller"]) then return false; end
	
	-- de-priorize pets
	if (a.pet and not b.pet) then return false; end;
	if (b.pet and not a.pet) then return true; end;
	
	-- CT RaidAssist support, if we have main tanks set in CT RaidAssist prefer them
	if (CT_RA_MainTanks ~= nil) then
		--SA_Debug("CT RaidAssist detected, searching for maintanks ..");
		local a_tank, b_tank = false;
		for _,v in CT_RA_MainTanks do
			if (a["unitName"] == v) then a_tank=true; end
			if (b["unitName"] == v) then b_tank=true; end
		end
		-- if both are CTRA tanks, sort by priority (mt first!)
		if (a_tank and b_tank) then 
			return SA_TableIndex(CT_RA_MainTanks, a.unitName) < SA_TableIndex(CT_RA_MainTanks, b.unitName); 
		end;
		
		if (a_tank) then return true; end
		if (b_tank) then return false; end
	end
	
	-- if we have same class, sort secondary by health
	if (a["class"] == b["class"]) then
		return a["health"] < b["health"];
	end
	
	-- and last, teh normal case where sorted by class
	return SA_TableIndex(SA_OPTIONS.ClassOrder, a["class"]) < SA_TableIndex(SA_OPTIONS.ClassOrder, b["class"]);
end

function SA_RefreshPuller()
	-- check that current puller exists in party, if not clear
	local candidates = SA_GetCandidates(true);
	if (not candidates[SA_OPTIONS["puller"]]) then
		SA_Debug("Puller does not exist in candidates, clearing");
		SA_OPTIONS["puller"]=nil;
	end
	
	-- if there is no puller and we have pet, set it as one
	if (SA_OPTIONS["puller"]==nil and UnitExists("pet")) then
		SA_Debug("no puller set, pet exists -> setting it as one");
		SA_OPTIONS["puller"] = UnitName("pet");
	end
end

function SA_WhoseTarget(unit, candidates)
	for k,candidate in candidates do
		if (UnitIsUnit(unit, candidate.target)) then
			return candidate;
		end
	end
	return nil;
end

-- params
-- allowNearest = allow fallback to target nearest
-- recursive = is this recursive call, if it is we cannot add previous assists "smartly" .. todo: better way?
function FindTarget(allowNearest, recursive)
	local candidates = SA_GetCandidates(false);

	-- if we have already good target selected (from outside smart assist), find out who we must skip (whos target is this?)
	-- this should NOT be done unless we have group (incase we have pet), update: not implemented but seems to work?
	-- it _should_ be targettarget (enemy's target) .. but  more reliable way is to iterate trough everyone and see whos 
	-- target it is
	if (UnitCanAttack("player", "target") and SA_TableSize(PREVIOUS_ASSISTS)==0 and not recursive) then
		SA_Debug("current target outside smartassist, finding if it's someones target");
		local who = SA_WhoseTarget("target", candidates);
		if (who) then
			SA_Debug("outside target is target of "..who.unitName.." adding to prev assists list!");
			PREVIOUS_ASSISTS[who["unitId"]] = true;
		end
	end

	--SA_DebugCandidates(candidates);
	
	SA_RefreshPuller();
	
	if (SA_TableSize(candidates)>1) then
		table.sort(candidates, function(a,b) return SA_SortCandidate(a,b) end);
		SA_DebugCandidates(candidates);
	end
	
	for _,candidate in candidates do
		-- this is ingenious loop which determines if current candidate target has been targetted on previous assists
		-- not a idiot proof check since previous party members might have changed target since then, but works amazingly well
		-- atleast preventing situation where multiple members have same target and you press assist key multiple times
		
		-- TODO PRIORITY: This doesn't still work 100% ... but something like 99%
		-- perhaps we could iterate everyone and get next who does not have same target as we do !
		
		local previously_targetted = false;
		for assisted,value in PREVIOUS_ASSISTS do
			if (UnitExists(assisted)) then
				if (UnitIsUnit(candidate.target, assisted.."target")) then
					SA_Debug("** already targetted once "..tostring(UnitName(assisted)));
					previously_targetted = true;
					break;
				end
			end
		end
		
		-- test each candidate, skips previously assisted UNLESS it has health below critical value
		if ( ((PREVIOUS_ASSISTS[candidate["unitId"]] == nil) or
			 (candidate["health"] < SA_OPTIONS.PriorizeHealthValue and SA_OPTIONS.PriorizeHealth)) and
			 (previously_targetted==false) ) 
		then
			-- test if candidate (partyN, pet, raidN etc) has valid target
			if(	UnitCanAssist("player", candidate["unitId"]) and 
				isValidTarget(candidate.target) ) 
			then
				if (SA_OPTIONS.VerboseAssistToSct and SCT_Display) then  
					SCT_Display("Assisting "..candidate.unitName, {r=0,g=0.8,b=1});
				end
				if (SA_OPTIONS.VerboseAssistToChat) then 
					printInfo("Assisting "..candidate.unitName);
				end
				SA_Debug("Found a good target from "..candidate["unitName"].." ("..candidate["unitId"]..").");
				AssistUnit(candidate["unitId"]);
				PREVIOUS_ASSISTS[candidate["unitId"]] = true;
				return;
			end
		else
			SA_Debug("** Skipping "..candidate["unitName"].." ("..candidate["unitId"]..")");
		end

	end
	-- if we have skiplist, now is good time to clear it and try again with all members..
	SA_Debug("** previous assist length: "..SA_TableSize(PREVIOUS_ASSISTS));
	if (SA_TableSize(PREVIOUS_ASSISTS)>0) then
		SA_Debug("** Unable to assist anyone but we have skiplist, clearing it and trying again..");
		PREVIOUS_ASSISTS = {};
		return FindTarget(allowNearest, true);
	end
	
	-- we might have some invalid (friendly) targetted by now
	ClearTarget();
	
	if (SA_OPTIONS.FallbackTargetNearest and allowNearest) then
		local members = SA_TableSize(candidates);
		if (SA_OPTIONS.DisableTargetNearest and members > SA_OPTIONS.DisableTargetNearestValue) then
			printInfo("Unable to assist anyone. Targetting nearest is suspended in groups this large.");
			return
		end
		printInfo("Unable to assist anyone. Trying to target nearest enemy.");
		--ClearTarget();
		TargetNearestEnemy();
		if (not isValidTarget("target")) then
			SA_Debug("Still no good target..");
			ClearTarget();
		end
	end	
end

-- Is the target valid for targetting
function isValidTarget(unit)
	--SA_Debug("isValidTarget "..unit);
	if (UnitExists(unit)) then
		-- Don't target dead units
		if (UnitIsDead(unit)) then
			SA_Debug("Dead, not targeting "..SA_TargetInfo(unit));
			return false;
		end
		-- Don't target friendlies
		if (UnitIsFriend(unit, "player")) then
			SA_Debug("Friendly, not targeting "..SA_TargetInfo(unit));
			return false;
		end
		-- Don't target if we can't attack it
        if (not UnitCanAttack(unit, "player")) then
        	SA_Debug("Can't attack, not targeting "..SA_TargetInfo(unit));
			return false;
        end
		-- Don't target CCd mobs
		if (isUnitCC(unit)) then
			SA_Debug("CCd, not targeting "..SA_TargetInfo(unit));
			return false;
		end
	else
		return false;
	end
	
	return true;
end

-- Check if the targeted unit has any CC methods applied to it (seduce, charm, polymorph)
-- return true if the unit is CCd
-- basic idea taken from pet leash. I've added many more spells to the list tough :)
function isUnitCC(unit)
	for i = 1, 10 do
		if (UnitDebuff(unit,i)) then
			currentDebuff = UnitDebuff(unit,i)
			--SA_Debug("currentDebuff "..currentDebuff);
			if (currentDebuff == "Interface\\Icons\\Spell_Shadow_Possession") then
				SA_Debug("-- Possessed, not targeting "..SA_TargetInfo(unit));
				return true;
			end
			if (currentDebuff == "Interface\\Icons\\Spell_Shadow_Charm") then
				SA_Debug("-- Charmed, not targeting "..SA_TargetInfo(unit));
				return true;
			end
			if (currentDebuff == "Interface\\Icons\\Spell_Nature_Polymorph") then
				SA_Debug("-- Polymorphed, not targeting "..SA_TargetInfo(unit));
				return true;
			end
			if (currentDebuff == "Interface\\Icons\\Spell_Frost_ChainsOfIce") then
				SA_Debug("-- Ice trapped, not targeting "..SA_TargetInfo(unit));
				return true;
			end
			if (currentDebuff == "Interface\\Icons\\Spell_Nature_Slow") then
				SA_Debug("-- Shackle Undead, not targeting "..SA_TargetInfo(unit));
				return true;
			end
			if (currentDebuff == "Interface\\Icons\\Spell_Nature_Sleep") then
				SA_Debug("-- Sleeped, not targeting "..SA_TargetInfo(unit));
				return true;
			end
			if (currentDebuff == "Interface\\Icons\\Ability_Sap") then
				SA_Debug("-- Sapped, not targeting "..SA_TargetInfo(unit));
				return true;
			end
			-- or is banish Spell_Shadow_CurseOfArcimonde ?
			if (currentDebuff == "Interface\\Icons\\Spell_Shadow_Cripple") then
				SA_Debug("-- Banished, not targeting "..SA_TargetInfo(unit));
				return true;
			end
		else
			-- no more debuffs on unit, stop the loop
			break;
		end
	end
	return false;
end

-- Return a descriptive string for the current target
function SA_TargetInfo(unit)
	name = UnitName(unit);
	level = UnitLevel(unit);
	return name
end

-- informational messages
function printInfo(message)
	DEFAULT_CHAT_FRAME:AddMessage(message, 0, 0.8, 1);
end

-- SA_Debug messages
function SA_Debug(message)
	if(SA_OPTIONS.Debug) then
		DEFAULT_CHAT_FRAME:AddMessage(message, 0, 0.8, 1);
	end
end

-------------------------------------------------------------------------------
-- Some table functions which I've been unable to find from lua
-------------------------------------------------------------------------------

function SA_TableSize(t)
	size = 0;
	for k,v in t do
		size = size + 1;
	end
	return size;
end

function SA_InList(list, value)
	for _,v in list do
		if (v==value) then return true; end
	end
	return false;
end

function SA_TableIndex(tbl, value)
	for k,v in tbl do
		if (v==value) then
			return k
		end	
	end
	return -1
end