NUM_SLOTS = 120;

-------------------------------------------------------------------------------
-- tab management code, credits to guim author
-------------------------------------------------------------------------------

SA_SUBFRAMES = { "SABasicOptionsFrame", "SAAdvancedOptionsFrame", "SmartAssistOptionsFrame" };

function SA_ToggleDialog(tab)
	local subFrame = getglobal(tab);
	if ( subFrame ) then
		PanelTemplates_SetTab(SAOptionsFrame, subFrame:GetID());
		if ( SAOptionsFrame:IsVisible() ) then
			PlaySound("igCharacterInfoTab");
			SA_ShowSubFrame(tab);
		else
			ShowUIPanel(SAOptionsFrame);
			SA_ShowSubFrame(tab);
		end
	end
end

function SA_ShowSubFrame(frameName)
	for index, value in SA_SUBFRAMES do
		if ( value == frameName ) then
			getglobal(value):Show()
		else
			getglobal(value):Hide();	
		end	
	end 
end

function SA_Tab_OnClick()
	SA_ToggleDialog(SA_SUBFRAMES[this:GetID()]);
	PlaySound("igCharacterInfoTab");
end

-------------------------------------------------------------------------------
-- spell & action drag hookups
-------------------------------------------------------------------------------

local SA_DRAGSPELL = {};
local SA_DRAGSLOT = -1;

function SA_ClearDrag()
	SA_DRAGSPELL = {};
	SA_DRAGSLOT = -1;
end
 
-- hook to spellbook pickup routine
local SA_BasePickupSpell = PickupSpell;
function PickupSpell(id, bookType)
	SA_ClearDrag();
	local result = SA_BasePickupSpell(id, bookType);
	--SA_Debug("picked up spell id="..id.." bookType="..bookType);
	if (CursorHasSpell()) then
		local name, rank = GetSpellName(id, bookType);
		local texture = GetSpellTexture(id, bookType);
		SA_DRAGSPELL.Id = id;
		SA_DRAGSPELL.Texture = texture;
		SA_DRAGSPELL.Name = name;
		SA_DRAGSPELL.BookType = bookType;
	end
	return result;
end

local SA_BasePickupAction = PickupAction;
function PickupAction(slot)
	SA_ClearDrag();
	SA_Debug("picked up spell from slot "..slot);
	-- set drag slot
	SA_DRAGSLOT = slot;
	return SA_BasePickupAction(slot);
end

-- detect slot where dragged, if we are dragging smart action update the id
local SA_PlaceAction = PlaceAction;
function PlaceAction(arg)
	SA_Debug("place action "..arg.. " dragslot="..SA_DRAGSLOT);

	local updated = false;
	-- smart action dragged, update the id
	for k,v in SA_OPTIONS.BuffSlot do
		if (v==SA_DRAGSLOT) then
			SA_Debug("*** updating buffslot "..SA_DRAGSLOT.." to "..arg);
			SA_OPTIONS.BuffSlot[k] = arg;
			updated = true;
			break;
		end
	end
	for k,v in SA_OPTIONS.AssistSlot do
		if (v==SA_DRAGSLOT) then
			SA_Debug("*** updating assistslot "..SA_DRAGSLOT.." to "..arg);
			SA_OPTIONS.AssistSlot[k] = arg;
			updated = true;
			break;
		end
	end

	-- if we dragged something else over the smart action it will get all screwed up, we need to remove it from the list
	-- in theory it MIGHT be possible to update the id but in reality the system gets so screwed up that I have no idea
	-- where it could be catched reliably. ie. on second spell the placeaction method is never called ... instead
	-- some other irrelevant methods are. go figure ..
	if (not updated) then
		local deleted = false;
		for k,v in SA_OPTIONS.BuffSlot do
			if (v==arg) then
				table.remove(SA_OPTIONS.BuffSlot, k);
				deleted = true;
				break;
			end
		end
		for k,v in SA_OPTIONS.AssistSlot do
			if (v==arg) then
				table.remove(SA_OPTIONS.AssistSlot, k);
				deleted = true;
				break;
			end
		end
		if (deleted) then
			SAMsgFrame:AddMessage("SmartAction lost, please reconfigure!", 1.0, 0, 0, 1.0, 7);
		end
	end	
	
	SA_ClearDrag();
	return SA_PlaceAction(arg);
end

-------------------------------------------------------------------------------
-- smart assist configuration dialog
-------------------------------------------------------------------------------

function SA_Options_Init()
end

function SA_Options_OnHide()
end

function SA_Options_OnLoad()
	-- make it draggable (?)
	UIPanelWindows['SmartAssistOptionsFrame'] = {area = 'center', pushable = 0};
	tinsert(UISpecialFrames, "SmartAssistOptionsFrame")
	-- Tab Handling code
	PanelTemplates_SetNumTabs(this, 3);
	PanelTemplates_SetTab(this, 1);
end

-- todo: this update is not anymore good since we have divided optiosn across several tabs ...
function SA_Options_OnShow()
	SAPriorizeHealthCB:SetChecked(SA_OPTIONS.PriorizeHealth);
	SAAssistOnEmoteCB:SetChecked(SA_OPTIONS.AssistOnEmote);
	SAVisualWarningCB:SetChecked(SA_OPTIONS.VisualWarning);
	SAAudioWarningCB:SetChecked(SA_OPTIONS.AudioWarning);
	SAAutoAssistCB:SetChecked(SA_OPTIONS.AutoAssist);
	SAFallbackTargetNearestCB:SetChecked(SA_OPTIONS.FallbackTargetNearest);

	SA_Options_UpdatePullerText();
		
	if (SA_OPTIONS.AutoAssistTexture) then
		SAAssistText:SetText(SA_OPTIONS.AutoAssistName);
		SA_Debug("setting button texture to="..SA_OPTIONS.AutoAssistTexture);
		SetItemButtonTexture(SAAssistWithSlot, SA_OPTIONS.AutoAssistTexture);
	end
	SA_Options_UpdateClassOrder();
	
	-- set health slider value & update text
	SAHealthSlider:SetValue(SA_OPTIONS.PriorizeHealthValue);
	SA_Options_UpdateHealthSlider();
	
	-- show / hide button text
	if (SA_OPTIONS.ShowAvailable) then
		SAToggleAvailable:SetText("Hide available");
	else
		SAToggleAvailable:SetText("Show available");
	end
	
end

function SA_Options_OnShowAdvanced()
	SAOutOfCombatCB:SetChecked(SA_OPTIONS.OutOfCombat);
	SADisableTargetNearestCB:SetChecked(SA_OPTIONS.DisableTargetNearest);
	
	-- update the dropdown Select Modifier
	UIDropDownMenu_SetSelectedID(SASelectModifierDD, SA_OPTIONS.AssistKeyMode);
	SASelectModifierDDText:SetText(SELECT_MODIFIER_TEXTS[SA_OPTIONS.AssistKeyMode]);
	if (SA_OPTIONS.AssistKeyMode==4) then
		SASelectModifierDDInfo:SetText("selecting does not assist");
	else
		SASelectModifierDDInfo:SetText("assists player");
	end
	-- update the dropdown Assist Modifier
	UIDropDownMenu_SetSelectedID(SAAssistModifierDD, SA_OPTIONS.DisableAutoCastKeyMode);
	SAAssistModifierDDText:SetText(ASSIST_MODIFIER_TEXTS[SA_OPTIONS.DisableAutoCastKeyMode]);

	
		
	-- set disable target nearest slider value & update text
	SADisableTargetNearest:SetValue(SA_OPTIONS.DisableTargetNearestValue);
	SA_Options_UpdateDisableTargetSlider();
		
	SAVerboseAssistToChatCB:SetChecked(SA_OPTIONS.VerboseAssistToChat);
	SAVerboseAssistToSctCB:SetChecked(SA_OPTIONS.VerboseAssistToSct);
	SAVerboseIncomingToSctCB:SetChecked(SA_OPTIONS.VerboseIncomingToSct);
	
	-- disable some checkboxes if sct is not available
	if (not SCT_Display) then
		SAVerboseAssistToSctCB:Disable();
	end
end


function SA_Options_UpdatePullerText()
	local candidates = SA_GetCandidates();
	SA_RefreshPuller(candidates);
	if (SA_OPTIONS["puller"]==nil) then
		SACurrentPuller:SetText("Assigned puller: none / pet when available");
	else
		SACurrentPuller:SetText("Assigned puller: "..SA_OPTIONS["puller"]);
	end
end

-- show the main configuration window
function SmartAssistOptions_Toggle()
	if (SmartAssistOptionsFrame:IsVisible()) then
		SmartAssistOptionsFrame:Hide();
	else
		SmartAssistOptionsFrame:Show();
	end
end

-- togle a boolean option, play a tick sound
function SA_ToggleOption(option)
	if (SA_OPTIONS[option]==nil or SA_OPTIONS[option]==false) then
		SA_OPTIONS[option] = true;
		PlaySound("igMainMenuOptionCheckBoxOff");
	else
		SA_OPTIONS[option] = false;
		PlaySound("igMainMenuOptionCheckBoxOn");
	end
end

function SA_ToggleAvailable()
	SA_ToggleOption("ShowAvailable");
	if (SA_OPTIONS.ShowAvailable) then
		printInfo("Available assist window is now ENABLED");
		SAListFrame:Show();
	else
		printInfo("Available assist window is now DISABLED");
		SAListFrame:Hide();
	end
	SA_Options_OnShow();
end

function SA_SpellSlot_OnReceiveDrag()
	SA_Debug("Received drag");
	
	if (CursorHasSpell() and SA_DRAGSLOT == -1) then
		SA_OPTIONS.AutoAssist = true;
		SA_OPTIONS.AutoAssistTexture = SA_DRAGSPELL.Texture;
		SA_OPTIONS.AutoAssistName = SA_DRAGSPELL.Name;

		-- drop icon
		PickupSpell(SA_DRAGSPELL.Id, SA_DRAGSPELL.BookType);
		
		-- update the config window		
		SA_Options_OnShow();
	else
		SpellBookFrame:Show();
	end
end

-- updates classorder view
function SA_Options_UpdateClassOrder()
	for k,v in SA_OPTIONS.ClassOrder do
		getglobal("ClassOrderFrameClass" .. k .. "Text"):SetText(v);
	end
end

function SA_Options_UpdateDisableTargetSlider()
	SA_OPTIONS.DisableTargetNearestValue = SADisableTargetNearest:GetValue();
	SADisableTargetNearestText:SetText("members > "..SA_OPTIONS.DisableTargetNearestValue);
end

-- updates health slider TEXT (not value)
function SA_Options_UpdateHealthSlider()
	SA_OPTIONS.PriorizeHealthValue = SAHealthSlider:GetValue();
	SAHealthSliderText:SetText("Priority health ("..SA_OPTIONS.PriorizeHealthValue.." %)");
end

-- class order routines
function SA_ClassOrderMove(move)
	local text = getglobal(this:GetParent():GetName() .. "Text"):GetText()
	
	-- where is our text in table, also check for boundaries
	local index = SA_TableIndex(SA_OPTIONS.ClassOrder, text);
	SA_Debug("index="..index.." move="..move);
	if (index==1 and move==-1) then
		SA_Debug("index is 1, and position is up, returning");
		return
	end
	if (index==9 and move==1) then
		SA_Debug("index is 9, and position is down, returning");
		return
	end

	-- move the element in our list	
	local moving = SA_OPTIONS.ClassOrder[index];
	table.remove(SA_OPTIONS.ClassOrder, index);
	table.insert(SA_OPTIONS.ClassOrder, index+move, moving);
	
	SA_Debug("text="..tostring( text ).."move from "..index.." to "..index+move);
	
	SA_Options_UpdateClassOrder();
end

-------------------------------------------------------------------------------
-- initialize dropdown (thanks to atlas)
-------------------------------------------------------------------------------

SELECT_MODIFIER_TEXTS = {"Shift - select", "Ctrl - select", "Alt - select", "Disabled" };
function SASelectModifierDD_Initialize()
	local info;
	for _,value in SELECT_MODIFIER_TEXTS do
		info = {
			text = value;
			func = SASelectModifierDDButton_OnClick;
		};
		UIDropDownMenu_AddButton(info);
	end
end

function SASelectModifierDD_OnLoad()
	UIDropDownMenu_Initialize(SASelectModifierDD, SASelectModifierDD_Initialize);
	UIDropDownMenu_SetSelectedID(SASelectModifierDD, 1);
	UIDropDownMenu_SetWidth(90);
end

function SASelectModifierDDButton_OnClick()
	local oldID = UIDropDownMenu_GetSelectedID(SASelectModifierDD);
	if(oldID ~= this:GetID()) then
		SA_Debug("SASelectModifierDD oldID="..oldID.." this:GetID()="..this:GetID());
		SA_OPTIONS["AssistKeyMode"]=this:GetID()
		SA_Options_OnShowAdvanced();
	end
end

-------------------------------------------------------------------------------

ASSIST_MODIFIER_TEXTS = {"Shift", "Ctrl ", "Alt", "None" };
function SAAssistModifierDD_Initialize()
	local info;
	for _,value in ASSIST_MODIFIER_TEXTS do
		info = {
			text = value;
			func = SAAssistModifierDDButton_OnClick;
		};
		UIDropDownMenu_AddButton(info);
	end
end

function SAAssistModifierDD_OnLoad()
	UIDropDownMenu_Initialize(SAAssistModifierDD, SAAssistModifierDD_Initialize);
	UIDropDownMenu_SetSelectedID(SAAssistModifierDD, 1);
	UIDropDownMenu_SetWidth(90);
end

function SAAssistModifierDDButton_OnClick()
	local oldID = UIDropDownMenu_GetSelectedID(SAAssistModifierDD);
	if(oldID ~= this:GetID()) then
		SA_Debug("SAAssistModifier oldID="..oldID.." this:GetID()="..this:GetID());
		SA_OPTIONS["DisableAutoCastKeyMode"]=this:GetID()
		SA_Options_OnShowAdvanced();
	end
end

-------------------------------------------------------------------------------
-- smart spell configuration methods
-------------------------------------------------------------------------------

function SP_Options_OnLoad()
	UIPanelWindows['SPOptionsFrame'] = {area = 'center', pushable = 0};
	tinsert(UISpecialFrames, "SPOptionsFrame")
end


function SA_RefreshSlots(name, list)
	-- clear old textures	
	for c = 1, 10 do
		local slot = getglobal(name..c);
		SetItemButtonTexture(slot, nil);
	end
	-- set textures
	i = 0;
	for _,v in list do
		i = i + 1;
		local slot = getglobal(name..i);
		SetItemButtonTexture(slot, GetActionTexture(v));
	end
end


function SP_Options_OnShow()
	if (SA_OPTIONS.AssistSlot == nil) then SA_OPTIONS.AssistSlot = {}; end;
	if (SA_OPTIONS.BuffSlot == nil) then SA_OPTIONS.BuffSlot = {}; end;
	
	SA_RefreshSlots("SPAssistSlot", SA_OPTIONS.AssistSlot);
	SA_RefreshSlots("SPBuffSlot", SA_OPTIONS.BuffSlot);
	
	-- refresh other widgets
	SPAfterBuffTargetBackMobCB:SetChecked(SA_OPTIONS.AfterBuffTargetBackMob);
	SPTriggerAssistCB:SetChecked(SA_OPTIONS.TriggerAssist);
	SPSmoothenBuffsCB:SetChecked(SA_OPTIONS.SmoothenBuffs);
	SPAutoLearnCB:SetChecked(SA_OPTIONS.AutoLearn);
end

function SP_Options_OnHide()
end

function SP_SpellSlot_OnClick()
	SA_Debug("SP_SpellSlot_OnClick, SA_DRAGSLOT = "..SA_DRAGSLOT);
	SA_Debug("slot CLICKED | name ="..tostring(this:GetName()).." id="..tostring(this:GetID()));
	
	if (SA_DRAGSLOT~=-1) then
		SP_SpellSlot_OnReceiveDrag();
		return;
	end
	
	if (this:GetID()>40) then
		-- if we have already something in there, remove it
		if (SA_OPTIONS.BuffSlot[this:GetID()-40]) then
			SA_Debug("buff slot has already stuff, removing it");
			table.remove(SA_OPTIONS.BuffSlot, this:GetID()-40);
		else
			--SP_SpellSlot_OnReceiveDrag();
			--return;
		end
	else
		-- if we have already something in there, remove it
		if (SA_OPTIONS.AssistSlot[this:GetID()]) then
			SA_Debug("assist slot has already stuff, removing it");
			table.remove(SA_OPTIONS.AssistSlot, this:GetID());
		else
			--SP_SpellSlot_OnReceiveDrag();
			--return;
		end
	end
	SP_Options_OnShow();
end

function SP_SpellSlot_OnReceiveDrag()
	--if (CursorHasSpell() and SA_DRAGSLOT == -1) then
	if (SA_DRAGSLOT == -1) then
		UIErrorsFrame:AddMessage("Drag from actionbars!", 1.0, 0, 0, 1.0, 3);		
	else
		SA_Debug("received slot "..SA_DRAGSLOT);
		if (this:GetID()>40) then
			-- if we have already something in there, remove it
			if (SA_OPTIONS.BuffSlot[this:GetID()-40]) then
				SA_Debug("buff slot has already stuff, removing from "..this:GetID()-40);
				table.remove(SA_OPTIONS.BuffSlot, this:GetID()-40);
			end
			SA_Debug("adding to buffs");
			table.insert(SA_OPTIONS.BuffSlot, SA_DRAGSLOT);
		else
			-- if we have already something in there, remove it
			if (SA_OPTIONS.AssistSlot[this:GetID()]) then
				SA_Debug("assist slot has already stuff, removing from "..this:GetID());
				table.remove(SA_OPTIONS.AssistSlot, this:GetID());
			end
			SA_Debug("adding to assists");
			table.insert(SA_OPTIONS.AssistSlot, SA_DRAGSLOT);
		end

		-- refresh the view
		SP_Options_OnShow();
		
		-- release the spell
		SA_Debug("releasing spell from slot "..SA_DRAGSLOT);
		PickupAction(SA_DRAGSLOT);
		SA_ClearDrag();
		--SA_BasePickupAction(SA_DRAGSLOT);
				
	end
end

-- reset all smartactions
function SA_ResetSmartActions()
	SA_OPTIONS.BuffSlot = {};
	SA_OPTIONS.AssistSlot = {};
	-- refresh the view
	SP_Options_OnShow();
end

-- try to autoconfigure smart actions
function SA_AutoConfigureSmartActions()
	local found = 0;
	for slot = 1, NUM_SLOTS do
		local spell = SA_GetSlotName(slot);
		if (spell) then
			if (AUTOCONF_ATTACKS[spell]) then
				if (SA_TableIndex(SA_OPTIONS.AssistSlot, slot) == -1) then 
					SA_Debug("found attack "..spell);
					table.insert(SA_OPTIONS.AssistSlot, slot);
					found = found + 1;
				end
			end
			if (AUTOCONF_BUFFS[spell]) then
				if (SA_TableIndex(SA_OPTIONS.BuffSlot, slot) == -1) then 
					SA_Debug("found buff "..spell);
					table.insert(SA_OPTIONS.BuffSlot, slot);
					found = found + 1;
				end
			end
		end
	end
	if (found>0) then
		printInfo("Auto configured "..found.." actions.");
		if (table.getn(SA_OPTIONS.BuffSlot)>10 or table.getn(SA_OPTIONS.AssistSlot)>10) then
			printInfo("Auto configure found more actions than it can show in options. All of them are configured but only 10 first are displayed.");
		end
	else
		printInfo("Auto configure was unable to find any actions. This is most likelly because your class does not (yet) have predefined set of spells or you are using non-english client.");
	end
	-- refresh the view
	SP_Options_OnShow();
end


-- find slot id by name, nil if not found
function SA_FindSlotByName(name)
	local textName;
	for slot = 1, NUM_SLOTS do
		SAActionTip:SetAction(slot);
		textName = SAActionTipTextLeft1:GetText();
		if (textName == name) then
			return slot;
		end		
	end
	return nil;
end

-- get name of spell in slot by id
function SA_GetSlotName(slot)
	SAActionTip:SetAction(slot);
	return SAActionTipTextLeft1:GetText();
end

