SA_LIST_BOXES = {};

-- todo: incombat interval ja outofcombat interval! performance++ =)
SA_REFRESHRATE = 1.5; -- live at somewhere else ? .. this is just to smoke out syncronization errors
SA_LASTREFRESH = SA_REFRESHRATE;

SA_ALERTCOL = "|cffff0000";
SA_TANKCOL = "|cff00ff00";

function SA_InitBar(bar, text)
	bar:SetMinMaxValues(0, 100);
	bar:SetValue(100);
	text:SetPoint("CENTER", bar:GetName(), "CENTER", 0, 0)
	SetTextStatusBarText(bar, text);
	ShowTextStatusBarText(bar);
end

function SA_List_Onload()
	-- initialize boxes
	for i=1, 10 do
		local box = {};
		box["frame"] = getglobal("Target"..i);
		box["mobText"] = getglobal("Target"..i.."MobText");
		box["targetText"] = getglobal("Target"..i.."TargetText");
		box["targetOf"] = getglobal("Target"..i.."TargetOf");
		box["mobBar"] = getglobal("Target"..i.."MobBar");
		box["targetBar"] = getglobal("Target"..i.."TargetBar");
		SA_LIST_BOXES[i] = box;
	end
end

function SA_List_OnUpdate(elapsed)
	SA_LASTREFRESH = SA_LASTREFRESH - elapsed;
	if (SA_LASTREFRESH > 0) then
		return;
	end
	SA_LASTREFRESH = SA_REFRESHRATE;
	SA_List_Update();
end

function SA_List_Target_OnClick(arg)
	SA_Debug("TargetClick");
	local target = arg.obj;
	if (not target) then printInfo("SmartAssist: Unknown target?"); return; end;
    -- if shift is down paste all targetters names (except tanks) to chat
    if (IsShiftKeyDown()) then
    	local names = "";
    	local i = 0;
    	for _,c in target.players do
    		if (not SA_IsTank(c.unitName)) then
    			names = names..c.unitName..", ";
    			i = i + 1;
    		end;
    	end
    	if (i>0) then 
    		names = string.sub(names, 0, -3);
	    	ChatFrameEditBox:SetText(names.." ");
			ChatFrameEditBox:Show();
    	end
    	return;
    end
	-- initiate assist
	SA_Debug("assisting "..target.players[1].unitName);
	AssistUnit(target.players[1].unitId);
	SA_PostAssist();
end

function SA_ToTextCol(r,g,b)
	return string.lower("|cff"..string.format("%.2X",255*r)..string.format("%.2X",255*g)..string.format("%.2X",255*b));
end

function SA_List_Target_GetTooltip(arg)
	local target = arg.obj;
	if (not target) then return "Unknown?"; end;
	local text = "Targetted by:\n";
	for k,v in target.players do
		-- colorize text by class
		local cv = RAID_CLASS_COLORS[string.upper(v.class)];
		local color="";
		if (not cv) then
			color = "|cff888888";
		else
			color = SA_ToTextCol(cv.r, cv.g, cv.b);
		end
		text = text .. color..v.unitName .. "|r" .. "\n";
	end
	return text;
end

function SA_List_Target_OnEnter(arg)
	local text = SA_List_Target_GetTooltip(arg);
	GameTooltip:SetOwner(arg, "ANCHOR_LEFT");
	GameTooltip:SetText(text,1,1,1,1,1);
	
	-- Add some "sleep" to refreshing the list. It's not good to update the list while user is trying to select something.
	SA_LASTREFRESH = SA_REFRESHRATE * 2;
end

function SA_List_Target_OnLeave(arg)
	GameTooltip:Hide();
end

function SA_List_Refresh()
	SA_Debug("Requesting refresh now, target="..tostring(UnitName("target")));
	SA_LASTREFRESH = 0;
end


-- updates the available assists list
local PREV_COUNT = -1;
local PREV_OVERLOADED = false;
local PREV_AGGRO_COUNT = 0;
local PREV_MOB_COUNT = 0;

 -- this table contains all variables used in incoming detection & printing
local incoming = { prev_iter=0, currently=0, prev_time=0 };

function SA_List_Update()
	-- if disabled, stop immediattely
	if (not SA_OPTIONS.ShowAvailable) then return; end
	
	local candidates = SA_GetCandidates(false);
	table.sort(candidates, function(a,b) return SA_SortCandidate(a,b) end); -- perhaps no sort at all?

	local aggro_count = 0;
	
	-- construct list of available assists / targets
	local targets = {};
	for _,candidate in candidates do
	
		if( UnitCanAssist("player", candidate.unitId) and 
		    UnitExists(candidate.target) and
		    ((UnitAffectingCombat(candidate.target) or SA_OPTIONS.OutOfCombat) and not UnitIsDead(candidate.target)) and 
		    UnitCanAttack("player", candidate.target)) 
		then
			local target = SA_GetExistingTarget(candidate, targets);
			if (target) then
				--SA_Debug("appending "..candidate.unitName.." to target "..target.unitName);
				table.insert(target.players, candidate);
				target.playersCount = target.playersCount + 1;
			else
				-- this candidate target is not targetted by anyone else, create new target to the list
				target = {};
				target["players"] = { candidate };
				target["name"] = UnitName(candidate.target);
				target["unitName"] = UnitName(candidate.target);
				target["health"] = UnitHealth(candidate.target);
				target["target"] = UnitName(candidate.target.."target");
				target["targetHealth"] = ceil( UnitHealth( candidate.target.."target" ) / UnitHealthMax( candidate.target.."target" ) * 100 );
				target["playersCount"] = 1;
				if (SA_IsMarked(candidate.target)) then
					target["marked"] = true;
				else
					target["marked"] = false;
				end
				-- set mine to true if this is my target
				target["mine"] = UnitIsUnit("target", target.players[1].target);
				
				table.insert(targets, target);
				--SA_Debug("adding new target "..target.unitName.." from "..candidate.unitName);
			end
		end
	end
	
	-- update the list
	local i = 0;
	local overloaded = false;
	
	for _,target in targets do
		i = i + 1;
		-- todo: smarter ...
		if (i>10) then 
			SA_Debug("over 10 mobs, aborting list ..");
			overloaded = true;
			break; 
		end;
		
		local box = SA_LIST_BOXES[i];
		box.frame.obj = target;
		box.frame:Show();
			
		-- set values & texts
		local targetText = target.players[1].unitName;
		if (target.playersCount > 1) then
			targetText = targetText.." + "..tostring(target.playersCount-1); -- -1 for the one who targets (who + n)
		end
		local mobText = target.name;
		if (string.len(mobText)>13) then
			mobText = string.sub(mobText, 0, 13) .. "...";
		end
		box.mobBar:SetValue(target.health);
		box.mobText:SetText(mobText);
		
		-- colorize box
		if (not UnitAffectingCombat(target.players[1].target)) then
			box.frame:SetBackdropColor(0,0,0,0.4);
			box.frame:SetBackdropBorderColor(0,0,0,0.4);
		else
			local addinc = true;
			-- set border color
			if (target.marked) then
				box.frame:SetBackdropBorderColor(1,0,0,1);
			elseif (target.mine) then
				box.frame:SetBackdropBorderColor(0,1,0,1);
			elseif (isUnitCC(target.players[1].target)) then
				box.frame:SetBackdropBorderColor(0,0,0,1);
				addinc = false;
			else
				box.frame:SetBackdropBorderColor(0.8,0.8,0.8,1);
			end
			-- add to incoming
			if (addinc) then
				incoming.currently = incoming.currently + 1;
			end
			
			-- set background color
			if (target.mine) then
				box.frame:SetBackdropColor(0,0.75,0,0.85);
			else
				box.frame:SetBackdropColor(0,0,0,0.5);
			end
		end
		
		-- colorize target target text in some situations
		if (target.target) then
			local col = "";
			local endcol = "";
			if (UnitName("player") == target.target) then
				-- targetting player
				aggro_count = aggro_count + 1;
				col = SA_ALERTCOL;
			elseif (SA_IsTank(target.target)) then
				-- targetting tank
				col = SA_TANKCOL;
			end
			if (col~="") then
				endcol = "|r";
			end
			box.targetBar:SetValue(target.targetHealth);
			box.targetText:SetText("T:"..col..target.target..endcol);
		else
			box.targetBar:SetValue(0);
			box.targetText:SetText("?");
		end
			
		box.targetOf:SetText(targetText);
		
	end
	-- if count has been changed update visibility
	if (i ~= PREV_COUNT) then
		SA_Debug("should hide box now!");
		for j = i + 1, PREV_COUNT do
			SA_Debug("hiding box, j="..j);
			if (j>10) then
				SA_Debug("aborting j>10 !");
				overloaded = true;
				break;
			end
			SA_LIST_BOXES[j].frame:Hide();
		end
	end
	
	-- if we have a gotten aggro, play optional soundaaaaaa to alert player
	if (aggro_count > PREV_AGGRO_COUNT) then
		if (SA_OPTIONS.AudioWarning) then PlaySound("AuctionWindowClose"); end;
		-- todo: make as option!
		if (SCT_Display) then SCT_Display("Acquired aggro!", {r=1,g=0,b=0}); end
	end
	PREV_AGGRO_COUNT = aggro_count;
	
	-- if we have new mobs in combat
	--SA_Debug("currently ="..incoming.currently.." previously="..incoming.prev_iter);
	if (SA_OPTIONS.VerboseIncomingToSct) then
		local now = time();
		if (incoming.currently > incoming.prev_iter and now - incoming.prev_time > 4) then
			if (SCT_Display) then
				local amount = incoming.currently - incoming.prev_iter;
				if (amount > 1) then
					SCT_Display("Incoming x"..amount.."+!", {r=1,g=0,b=0}); 
				else
					SCT_Display("Incoming!", {r=1,g=0,b=0}); 
				end
			end
			incoming.prev_time = now;
		end
		incoming.prev_iter = incoming.currently;
		incoming.currently = 0;
	end
	
	if (overloaded and not PREV_OVERLOADED) then
		SAListFrame:SetBackdropColor(1,0,0,1);
		SAListFrame:SetBackdropBorderColor(1,0,0,1)
		SAListTitle:SetText("OVERLOADED");
	elseif (not overloaded and PREV_OVERLOADED) then
		SA_List_Init();
	end
	
	PREV_COUNT = i;
	PREV_OVERLOADED = overloaded;
end

function SA_List_Init()
	SAListFrame:SetBackdropColor(0, 0, 0, 0.3);
	SAListFrame:SetBackdropBorderColor(1,1,1,1)
	SAListTitle:SetText("Available assists");
end

-- return true if unit is marked with hunter's mark
function SA_IsMarked(unit)
	for i = 1, 10 do
		if (UnitDebuff(unit, i)) then
			currentDebuff = UnitDebuff(unit,i)
			if (currentDebuff == "Interface\\Icons\\Ability_Hunter_SniperShot") then return true; end
		else
			-- no more debuffs on unit, stop the loop
			break;
		end
	end
	return false;
end

-- return true if unitName is tank (in smartassist or in ct raidassist)
function SA_IsTank(unitName)
	if (unitName == SA_OPTIONS.puller) then return true; end;
	if (CT_RA_MainTanks) then
		for _,v in CT_RA_MainTanks do
			if (unitName == v) then return true; end
		end
	end
	return false;
end

-- check if candidates target is already targetted by someone, return the target if found and nil if not
function SA_GetExistingTarget(candidate, targets)
	for _,target in targets do
		if (UnitIsUnit(target.players[1].target, candidate.target)) then
			return target;
		end
	end
	return nil;
end


function SA_List_OnEnter()
	--this:SetBackdropColor(1,1,1,1);
	--this:SetBackdropBorderColor(1,1,1,1);
end

function SA_List_OnLeave()
	--this:SetBackdropColor(0,0,0,0);
	--this:SetBackdropBorderColor(0,0,0,0);
end